/*
 * The auptrace program uses the ptrace system call to follow forks
 * while tracing.  It adds audit rules as ptrace reports new descendants.
 * Copyright (C) 2006 The MITRE Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Jeremy Latt figured out how to use ptrace from manual pages and
   wrote an initial version of this code in a mixture of Python and C.
   John D. Ramsdell translated the program into C, and fixed a few
   bugs, including one in the ptrace algorithm. */

/* When run in debug mode, the program prints signals, but does not
   interact with the audit system.  This is useful when there are
   debugging this program. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <search.h>
#include <libaudit.h>
#include <sys/ptrace.h>
#ifdef HAVE_LINUX_PTRACE_H
#include <linux/ptrace.h>
#endif

/* Signal handler used to route a signal to the parent to its traced
   children.  The handler removes itself after one use. */

static pid_t group = 0;		/* Process group ID */

static void
signal_group(int signum)
{
  if (group && kill(-group, signum))
    perror("kill group");
  if (signal(signum, SIG_DFL) == SIG_ERR)
    fprintf(stderr, "Setting signal %d to its default failed\n", signum);
}

/* Constants taken from gdb/linux-nat.c. */

/* If the system headers did not provide the constants, hard-code the normal
   values.  */
#ifndef PTRACE_EVENT_FORK

#define PTRACE_SETOPTIONS	0x4200
#define PTRACE_GETEVENTMSG	0x4201

/* options set using PTRACE_SETOPTIONS */
#define PTRACE_O_TRACESYSGOOD	0x00000001
#define PTRACE_O_TRACEFORK	0x00000002
#define PTRACE_O_TRACEVFORK	0x00000004
#define PTRACE_O_TRACECLONE	0x00000008
#define PTRACE_O_TRACEEXEC	0x00000010
#define PTRACE_O_TRACEVFORKDONE	0x00000020
#define PTRACE_O_TRACEEXIT	0x00000040

/* Wait extended result codes for the above trace options.  */
#define PTRACE_EVENT_FORK	1
#define PTRACE_EVENT_VFORK	2
#define PTRACE_EVENT_CLONE	3
#define PTRACE_EVENT_EXEC	4
#define PTRACE_EVENT_VFORK_DONE	5
#define PTRACE_EVENT_EXIT	6

#endif /* PTRACE_EVENT_FORK */

/* Audit control functions */

#define AUPTRACE_FLAGS AUDIT_FILTER_ENTRY
#define AUPTRACE_ACTION AUDIT_ALWAYS
#define MAX_CHILDREN (1 << 10)
#define CHILD_DIGIT_SIZE 12
#define PID_FIELD_SIZE (4 + CHILD_DIGIT_SIZE)

static void *
xmalloc(size_t size)
{
  void *p = malloc(size);
  if (!p) {
    fprintf(stderr, "Failed to allocate space\n");
    exit(1);
  }
  return p;
}

/* Storage for traced children */
static int nchildren;
static int max_children;
static pid_t *children;

/* Debug mode flag */
static int debug = 0;

static int
init_child_store(void)
{
  children = xmalloc(max_children * sizeof(*children));
  nchildren = 0;
  int rc = !hcreate(max_children + (max_children >> 3));
  if (rc)
    fprintf(stderr, 
	    "Failed to create hash table for %d children\n", 
	    max_children);
  return rc;
}

static int
find_child(pid_t pid)		/* Returns true if child present */
{
  char key[CHILD_DIGIT_SIZE];
  ENTRY item = {key, NULL};
  if (snprintf(key, CHILD_DIGIT_SIZE, "%d", pid) >= CHILD_DIGIT_SIZE) {
    fprintf(stderr, "Internal error in add_child\n");
    exit(1);
  }
  return NULL != hsearch(item, FIND);
}

static int
add_child(pid_t pid)		/* Assumes nchildren < max_children */
{
  char *key = xmalloc(CHILD_DIGIT_SIZE);
  ENTRY item = {key, key};
  if (snprintf(key, CHILD_DIGIT_SIZE, "%d", pid) >= CHILD_DIGIT_SIZE) {
    fprintf(stderr, "Internal error in add_child\n");
    exit(1);
  }
  if (!hsearch(item, ENTER)) {
    fprintf(stderr, "Hash table full\n");
    return 1;
  }
  children[nchildren++] = pid;
  return 0;
}

static int audit_sys; /* File descriptor for socket to audit system */

static struct audit_rule_data *
init_rule(pid_t pid)		/* Create a rule for a pid */
{				/* Returns NULL on failure */
  char field[PID_FIELD_SIZE];
  if (snprintf(field, PID_FIELD_SIZE, "pid=%d", pid) >= PID_FIELD_SIZE) {
    fprintf(stderr, "Internal error in init_rule\n");
    exit(1);
  }
  struct audit_rule_data *r = xmalloc(sizeof(*r));
  memset(r, 0, sizeof(*r));
  if (audit_rule_syscallbyname_data(r, "all") < 0) {
    fprintf(stderr, "Illegal syscall name for audit rule for %d\n", pid);
    return NULL;
  }
  if (audit_rule_fieldpair_data(&r, field, AUPTRACE_FLAGS)) {
    fprintf(stderr, "Cannot add field to audit rule for %d\n", pid);
    return NULL;
  }
  return r;
}

/* A dummy implementations of add_rule */
static int
debug_add_rule(pid_t pid)
{
  if (find_child(pid)) {	/* Rule already added */
    fprintf(stderr, "Avoiding adding duplicate rule for %d\n", pid);
    return 0;
  }
  if (nchildren >= MAX_CHILDREN) {
    fprintf(stderr, "Too many children\n");
    return 1;
  }
  fprintf(stderr, "If not debugging, would be adding rule for %d\n", pid);
  return add_child(pid);
}

/* Equivalent to '/sbin/auditctl -a entry,always -F pid=%d -S all' */
static int
add_rule(pid_t pid)		/* Returns zero on success */
{
  if (debug)
    return debug_add_rule(pid);
  if (find_child(pid))		/* Rule already added */
    return 0;
  if (nchildren >= MAX_CHILDREN) {
    fprintf(stderr, "Too many children\n");
    return 1;
  }
  struct audit_rule_data *r = init_rule(pid);
  if (!r)
    return 1;
  if (audit_add_rule_data(audit_sys, r, AUPTRACE_FLAGS, AUPTRACE_ACTION) <= 0) {
    fprintf(stderr, "Cannot add an audit rule for %d\n", pid);
    return 1;
  }
  free(r);
  return add_child(pid);
}

/* Equivalent to '/sbin/auditctl -d entry,always -F pid=%d -S all' for
   every traced descendant. */
static void
delete_rules(void)
{
  int i;
  for (i = 0; i < nchildren; i++) {
    pid_t pid = children[i];
    struct audit_rule_data *r = init_rule(pid);
    if (!r)
      fprintf(stderr, "Cannot delete an audit rule for %d\n", pid);
    int rc = audit_delete_rule_data(audit_sys, r, AUPTRACE_FLAGS, 
				    AUPTRACE_ACTION);
    free(r);
    if (rc <= 0) 
      fprintf(stderr, "Cannot delete an audit rule for %d\n", pid);
  }
}

static int
geteventmsg(int pid, unsigned long *msg)
{
  int rc = ptrace(PTRACE_GETEVENTMSG, pid, NULL, msg);
  if (debug && rc < 0)
    fprintf(stderr, 
	    "Bad result for ptrace(PTRACE_GETEVENTMSG, %d, ...)\n",
	    pid);
  return rc;
}

static int
setoptions(int pid, int options)
{
  int rc = ptrace(PTRACE_SETOPTIONS, pid, NULL, options);
  if (debug && rc < 0)
    fprintf(stderr, 
	    "Bad result for ptrace(PTRACE_SETOPTIONS, %d, ..., %d)\n",
	    pid, options);
  return rc;
}

static int
restart(int pid, int signal)
{
  int rc = ptrace(PTRACE_CONT, pid, NULL, signal);
  if (debug && rc < 0)
    fprintf(stderr, 
	    "Bad result for ptrace(PTRACE_CONT, %d, ..., %d)\n",
	    pid, signal);
  return rc;
}

static int
traceme(void)
{
  int rc = ptrace(PTRACE_TRACEME, 0, NULL, NULL);
  if (debug && rc < 0)
    fprintf(stderr, "Bad result for ptrace(PTRACE_TRACEM, ...)\n");
  return rc;
}

static int			/* Wait but ignore interrupted */
wait_for_it(int *status)	/* system calls */
{
  int pid;
  do {
    pid = wait(status);
  } 
  while (pid < 0 && errno == EINTR);
  return pid;
}
 
static int			/* Watch all children */
watch(pid_t pid)		/* This process' child is pid  */
{				/* Function returns an exit code */
  if (add_rule(pid))
    return 1;
  for (;;) {
    int status;
    pid = wait_for_it(&status);
    if (pid < 0) {
      if (errno == ECHILD)	/* No children to wait for */
	return 0;		/* Declare success */
      perror("wait");
      return 1;
    }
    if (debug)
      fprintf(stderr, "Wait status for %d is %d (0x%x)\n",
	      pid, status, status);
    if (WIFSTOPPED(status)) {
      int signal = WSTOPSIG(status);
      if (debug)
	fprintf(stderr, "Process %d stopped with signal = %d\n", 
		pid, signal);
      if (signal == SIGTRAP) {	/* Tracing caused this signal */
	signal = SIGSTOP;
	unsigned long msg;
	if (geteventmsg(pid, &msg) < 0) {
	  perror("geteventmsg: ptrace(PTRACE_GETEVENTMSG, ...)");
	  return 1;
	}
	pid_t child = (pid_t)msg;
	if (child) {
	  /* The child of each traced fork is noted here */
	  if (debug)
	    fprintf(stderr, "Process %d forked %d\n", pid, child);
	  if (add_rule(child))
	    return 1;
	}
	else {
	  if (debug)
	    fprintf(stderr, 
		    "Set options on %d due to SIGTRAP without a child\n", 
		    pid);
	  if (setoptions(pid, PTRACE_O_TRACEFORK) < 0) {
	    perror("setoptions: ptrace(PTRACE_SETOPTIONS, ...)");
	    return 1;
	  }
	}
      }
      if (restart(pid, signal) < 0) {
	perror("restart: ptrace(PTRACE_CONT, ...)");
	return 1;
      }
    }
  }
}

static int 			/* Behavior on the parent */
parent(pid_t pid)		/* Returns an exit code */
{
  group = pid;			/* Child ID is the group id */
  if (signal(SIGINT, signal_group) == SIG_ERR) /* Send signal to group */
    fprintf(stderr, "Setting signal %d failed\n", SIGINT); /* once */

  int rc = watch(pid);
  if (debug) {
    fprintf(stderr, "Found %d descendants\n", nchildren);
    return rc;
  }
  delete_rules();
  printf("\nAudit records can be retrieved with:\n\n");
  int i;
  for (i = 0; i < nchildren; i++)
    printf("ausearch -r -p %d > %10.10d.txt\n", children[i], children[i]);
  return rc;
}

static int
set_id(const char *username)	/* Sets the group and user ID */
{				/* when given a user name. */
  struct passwd *pent;		/* Setuid and setgid bits are */
  if (!username)		/* ignored on the executable file */
    return 0;			/* containing the program being */
  else if ((pent = getpwnam(username)) == NULL) { /* traced. */
    fprintf(stderr, "Cannot find user `%s'\n", username);
    return -1;
  }
  else if (setregid(pent->pw_gid, pent->pw_gid)) {
    perror("setregid");
    return -1;
  }
  else if (setreuid(pent->pw_uid, pent->pw_uid)) {
    perror("setreuid");
    return -1;
  }
  else
    return 0;
}

static char *username = NULL;

static int 
trace(char **argv)		/* Set up first child for tracing. */
{				/* Returns an exit code */
  pid_t pid = fork();
  switch (pid) {
  case -1:			/* Error */
    perror("fork failed");
    return -1;
  case 0:			/* Child */
    if (set_id(username))
      return 1;
    if (setpgid(0, 0)) {	/* Make child a process group leader */
      perror("setpgid of child"); /* so that signals can be sent to */
      return 1;			  /* the group */
    }
    if (traceme() < 0) {
      perror("ptrace(PTRACE_TRACEME, ...)");
      return 1;
    }
    execvp(argv[0], argv);
    perror("exec");
    return 1;
  default:			/* Parent */
    return parent(pid);
  }
}

static void
usage(const char *prog)
{
  fprintf(stderr,
	  "Usage: %s [options] command...]\n"
	  "Options:\n"
	  "  -o file   -- output to file (default is standard output)\n"
	  "  -n digits -- set maximun number of childern (default %d)\n"
	  "  -u user   -- run command as user\n"
	  "  -d        -- run in debug mode (show signals but doesn't audit)\n"
	  "  -h        -- print this message\n"
          "Note: one interrupt from the keyboard is relayed to the children.\n",
	  prog, MAX_CHILDREN);
#if defined PACKAGE_STRING
  fprintf(stderr, "Package: %s\n", PACKAGE_STRING);
#endif
}

int main(int argc, char **argv)
{
  extern char *optarg;
  extern int optind;

  char *output = 0;
  max_children = MAX_CHILDREN;
  username = NULL;
  debug = 0;
  group = 0;

  for (;;) {
    int c = getopt(argc, argv, "o:n:u:dh");
    if (c == -1)
      break;
    switch (c) {
    case 'o':
      output = optarg;
      break;
    case 'n':
      max_children = atoi(optarg);
      if (max_children <= 0) {
	fprintf(stderr, "Bad maximun children argument %d\n", max_children);
	usage(argv[0]);
	return 1;
      }
      if (max_children < MAX_CHILDREN)
	max_children = MAX_CHILDREN;
      break;
    case 'u':
      username = optarg;
      break;
    case 'd':
      debug = 1;
      break;
    case 'h':
      usage(argv[0]);
      return 0;
    default:
      usage(argv[0]);
      return 1;
    }
  }

  if (argc == optind) {
    fprintf(stderr, "Bad argument count\n");
    usage(argv[0]);
    return 1;
  }

  if (init_child_store())
    return 1;

  if (!debug) {			/* Set up audit rules */

    if (getuid()) {
      fprintf(stderr, "You must be root to run this program\n");
      return 1;
    }

    set_aumessage_mode(MSG_STDERR, DBG_NO);

    audit_sys = audit_open();
    if (audit_sys < 0) {
      fprintf(stderr, "Connecting to the audit system failed\n");
      return 1;
    }

  }
  else			   /* In debug mode, don't add audit rules. */
    if (username && getuid()) {
      fprintf(stderr, "You must be root to run this program as %s\n", 
	      username);
      return 1;
    }
  
  if (output && !freopen(output, "w", stdout)) {
    perror(output);
    return 1;
  }

  return trace(argv + optind);
}
